# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.utils import build_engine_key_prefix
from myreco.engines.base_model import EnginesModelBase
from myreco.utils import extend_swagger_json
from copy import deepcopy
import asyncio


class EnginesDataImporterModelBase(EnginesModelBase):
    __swagger_json__ = extend_swagger_json(
        EnginesModelBase.__swagger_json__,
        __file__
    )

    @classmethod
    async def post_import_data_job(cls, req, session):
        session = cls._copy_session(session)
        engine = await cls._get_engine(req, session)
        if engine is None:
            return cls._build_response(404)

        jobs_id = cls._get_jobs_id_importer(engine)
        response = cls._create_job(cls._run_import_data_job, jobs_id, req, session, engine)
        return response

    @classmethod
    async def _get_engine(cls, req, session):
        id_ = req.path_params['id']
        engines = await cls.get(session, {'id': id_}, todict=False)

        if not engines:
            return None

        return engines[0]

    @classmethod
    def _get_jobs_id_importer(cls, engine):
        return cls._get_jobs_id(engine) + '_importer'

    @classmethod
    def _get_jobs_id(cls, engine):
        return build_engine_key_prefix({'id': engine.id, 'core': {'name': engine.core.name}})

    @classmethod
    def _run_import_data_job(cls, req, session, engine):
        return engine.core_instance.get_data(session)

    @classmethod
    async def get_import_data_job(cls, req, session):
        engine = await cls._get_engine(req, session)
        if engine is None:
            return cls._build_response(404)

        jobs_id = cls._get_jobs_id_importer(engine)
        return await cls._get_job(jobs_id, req, session)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.users.models import (GrantsModelBase, URIsModelBase, MethodsModelBase,
    UsersModelBase, build_users_grants_table, build_users_stores_table)
from myreco.stores.model import StoresModelBase
from myreco.external_variables.model import ExternalVariablesModelBase
from myreco.placements.models import (PlacementsModelBase, VariationsModelBase,
    ABTestUsersModelBase, build_variations_slots_table)
from myreco.slots.models import (SlotsVariablesModelBase, SlotsFiltersModelBase,
    SlotsModelBase, build_slots_fallbacks_table)
from myreco.engines.base_model import EnginesModelBase
from myreco.engines.data_importer.model import EnginesDataImporterModelBase
from myreco.engines.objects_exporter.model import EnginesObjectsExporterModelBase
from myreco.engines.cores.model import EnginesCoresModelBase
from myreco.items_types.model import ItemsTypesModelBase, build_items_types_stores_table
from myreco.items_types.data_file_importer.model import ItemsTypesDataFileImporterModelBase
from myreco.items_types.filters_updater.model import ItemsTypesFiltersUpdaterModelBase
from swaggerit.models.orm.factory import FactoryOrmModels


class FactoryError(Exception):
    pass


class ModelsFactory(object):

    def __init__(self, commons_models_attributes=None, commons_tables_attributes=None):
        self.base_model = FactoryOrmModels.make_sqlalchemy_redis_base()
        self.meta_class = type(self.base_model)
        self._commons_models_attrs = self._init_attributes(commons_models_attributes)
        self._commons_tables_attrs = self._init_attributes(commons_tables_attributes)

    def _init_attributes(self, attrs=None, update=None):
        if attrs is None:
            if update is None:
                attrs = dict()
            else:
                attrs = update
        elif update is not None:
            attrs.update(update)
        return attrs

    def make_all_models(self, app_type='recommender',
                        engines_recommender_base=EnginesModelBase,
                        engines_data_importer_base=EnginesDataImporterModelBase,
                        engines_objects_exporter_base=EnginesObjectsExporterModelBase,
                        engines_core_base=EnginesCoresModelBase,
                        slots_base=SlotsModelBase,
                        slots_variables_base=SlotsVariablesModelBase,
                        slots_filters_base=SlotsFiltersModelBase,
                        items_types_recommender_base=ItemsTypesModelBase,
                        items_types_data_importer_base=ItemsTypesDataFileImporterModelBase,
                        items_types_objects_exporter_base=ItemsTypesFiltersUpdaterModelBase,
                        placements_base=PlacementsModelBase,
                        variations_base=VariationsModelBase,
                        ab_test_users_base=ABTestUsersModelBase,
                        stores_base=StoresModelBase,
                        uris_base=URIsModelBase,
                        users_base=UsersModelBase,
                        methods_base=MethodsModelBase,
                        grants_base=GrantsModelBase,
                        external_variables_base=ExternalVariablesModelBase):
        self.make_all_tables()
        app_types = {'recommender', 'data_importer', 'objects_exporter'}
        if (app_type not in app_types):
            raise FactoryError(
                "Invalid application type '{}'. Valid types: {}".format(
                    app_type, ', '.join(app_types)))

        return {
            'engines': self.make_engines_model(
                app_type,
                recommender_base=engines_recommender_base,
                data_importer_base=engines_data_importer_base,
                objects_exporter_base=engines_objects_exporter_base
            ),
            'engines_cores': self.make_engines_cores_model(engines_core_base),
            'slots': self.make_slots_model(slots_base),
            'slots_variables': self.make_slots_variables_model(slots_variables_base),
            'slots_filters': self.make_slots_filters_model(slots_filters_base),
            'items_types': self.make_items_types_model(
                app_type,
                recommender_base=items_types_recommender_base,
                data_importer_base=items_types_data_importer_base,
                objects_exporter_base=items_types_objects_exporter_base
            ),
            'placements': self.make_placements_model(placements_base),
            'variations': self.make_variations_model(variations_base),
            'ab_test_users': self.make_ab_test_users_model(ab_test_users_base),
            'stores': self.make_stores_model(stores_base),
            'uris': self.make_uris_model(uris_base),
            'users': self.make_users_model(users_base),
            'methods': self.make_methods_model(methods_base),
            'grants': self.make_grants_model(grants_base),
            'external_variables': self.make_external_variables_model(external_variables_base)
        }

    def make_all_tables(self):
        return {
            'users_grants': self.make_users_grants_table(),
            'users_stores': self.make_users_stores_table(),
            'variations_slots': self.make_variations_slots_table(),
            'slots_fallbacks': self.make_slots_fallbacks_table(),
            'items_types_stores': self.make_items_types_stores_table()
        }

    def make_users_grants_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_users_grants_table(self.base_model.metadata, **attributes)

    def make_users_stores_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_users_stores_table(self.base_model.metadata, **attributes)

    def make_variations_slots_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_variations_slots_table(self.base_model.metadata, **attributes)

    def make_slots_fallbacks_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_slots_fallbacks_table(self.base_model.metadata, **attributes)

    def make_items_types_stores_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_items_types_stores_table(self.base_model.metadata, **attributes)

    def make_engines_model(self, app_type,
                           recommender_base=EnginesModelBase,
                           data_importer_base=EnginesDataImporterModelBase,
                           objects_exporter_base=EnginesObjectsExporterModelBase,
                           attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)

        if app_type == 'recommender':
            bases_classes = (recommender_base, self.base_model)
        elif app_type == 'data_importer':
            bases_classes = (data_importer_base, self.base_model)
        elif app_type == 'objects_exporter':
            bases_classes = (objects_exporter_base, self.base_model)

        return self.meta_class('EnginesModel', bases_classes, attributes)

    def make_engines_cores_model(self, base=EnginesCoresModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'EnginesCoresModel', (base, self.base_model), attributes)

    def make_slots_model(self, base=SlotsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'SlotsModel', (base, self.base_model), attributes)

    def make_slots_variables_model(self, base=SlotsVariablesModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('SlotsVariablesModel', (base, self.base_model), attributes)

    def make_slots_filters_model(self, base=SlotsFiltersModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('SlotsFiltersModel', (base, self.base_model), attributes)

    def make_items_types_model(self, app_type,
                               recommender_base=ItemsTypesModelBase,
                               data_importer_base=ItemsTypesDataFileImporterModelBase,
                               objects_exporter_base=ItemsTypesFiltersUpdaterModelBase,
                               attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)

        if app_type == 'recommender':
            bases_classes = (recommender_base, self.base_model)
        elif app_type == 'data_importer':
            bases_classes = (data_importer_base, self.base_model)
        elif app_type == 'objects_exporter':
            bases_classes = (objects_exporter_base, self.base_model)

        return self.meta_class('ItemsTypesModel', bases_classes, attributes)

    def make_placements_model(self, base=PlacementsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'PlacementsModel', (base, self.base_model), attributes)

    def make_variations_model(self, base=VariationsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'VariationsModel', (base, self.base_model), attributes)

    def make_ab_test_users_model(self, base=ABTestUsersModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'ABTestUsersModel', (base, self.base_model), attributes)

    def make_stores_model(self, base=StoresModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'StoresModel', (base, self.base_model), attributes)

    def make_uris_model(self, base=URIsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'URIsModel', (base, self.base_model), attributes)

    def make_users_model(self, base=UsersModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'UsersModel', (base, self.base_model), attributes)

    def make_methods_model(self, base=MethodsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'MethodsModel', (base, self.base_model), attributes)

    def make_grants_model(self, base=GrantsModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'GrantsModel', (base, self.base_model), attributes)

    def make_external_variables_model(self, base=ExternalVariablesModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'ExternalVariablesModel', (base, self.base_model), attributes)


# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.filters.factory import FiltersFactory
from swaggerit.utils import get_swagger_json
from swaggerit.exceptions import SwaggerItModelError
from jsonschema import ValidationError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
import sqlalchemy as sa
import ujson


class SlotsVariablesModelBase(AbstractConcreteBase):
    __tablename__ = 'slots_variables'
    __use_redis__ = False
    __id_names__ = ['id']

    # To autoincrement works a alter table with autoincrement is necessary
    id = sa.Column(sa.Integer, nullable=False, unique=True, autoincrement=True, index=True)
    override = sa.Column(sa.Boolean, default=False)
    override_value_json = sa.Column(sa.Text)
    engine_variable_name = sa.Column(sa.String(255), primary_key=True)

    @property
    def override_value(self):
        if not hasattr(self, '_override_value'):
            self._override_value = \
                ujson.loads(self.override_value_json) if self.override_value_json is not None else None
        return self._override_value

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'override_value':
            value = ujson.dumps(value)
            attr_name = 'override_value_json'

        await super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('override_value') is not False:
            dict_inst.pop('override_value_json')
            dict_inst['override_value'] = self.override_value

    @declared_attr
    def external_variable_name(cls):
        return sa.Column(sa.ForeignKey('external_variables.name', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def external_variable_store_id(cls):
        return sa.Column(sa.ForeignKey('external_variables.store_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def slot_id(cls):
        return sa.Column(sa.ForeignKey('slots.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True)

    @declared_attr
    def external_variable(cls):
        return sa.orm.relationship('ExternalVariablesModel',
            foreign_keys=[cls.external_variable_name, cls.external_variable_store_id],
            primaryjoin='and_(SlotsVariablesModel.external_variable_name == ExternalVariablesModel.name, '\
                        'SlotsVariablesModel.external_variable_store_id == ExternalVariablesModel.store_id)')


class SlotsFiltersModelBase(AbstractConcreteBase):
    __tablename__ = 'slots_filters'
    __use_redis__ = False
    __factory__ = FiltersFactory
    __id_names__ = ['id']

    # To autoincrement works a alter table with autoincrement is necessary
    id = sa.Column(sa.Integer, nullable=False, unique=True, autoincrement=True, index=True)
    is_inclusive = sa.Column(sa.Boolean, default=True, primary_key=True)
    property_name = sa.Column(sa.String(255), nullable=False, primary_key=True)
    type_id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    override = sa.Column(sa.Boolean, default=False)
    override_value_json = sa.Column(sa.Text)
    skip_values_json = sa.Column(sa.Text)

    @property
    def override_value(self):
        if not hasattr(self, '_override_value'):
            self._override_value = \
                ujson.loads(self.override_value_json) if self.override_value_json is not None else None
        return self._override_value

    @property
    def skip_values(self):
        if not hasattr(self, '_skip_values'):
            self._skip_values = \
                ujson.loads(self.skip_values_json) if self.skip_values_json is not None else None
        return self._skip_values

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'skip_values':
            value = ujson.dumps(value)
            attr_name = 'skip_values_json'

        if attr_name == 'override_value':
            value = ujson.dumps(value)
            attr_name = 'override_value_json'

        await super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('skip_values') is not False:
            dict_inst.pop('skip_values_json')
            dict_inst['skip_values'] = self.skip_values

        if schema.get('override_value') is not False:
            dict_inst.pop('override_value_json')
            dict_inst['override_value'] = self.override_value

    @declared_attr
    def external_variable_name(cls):
        return sa.Column(sa.ForeignKey('external_variables.name', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def external_variable_store_id(cls):
        return sa.Column(sa.ForeignKey('external_variables.store_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def slot_id(cls):
        return sa.Column(sa.ForeignKey('slots.id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def external_variable(cls):
        return sa.orm.relationship('ExternalVariablesModel',
            foreign_keys=[cls.external_variable_name, cls.external_variable_store_id],
            primaryjoin='and_(SlotsFiltersModel.external_variable_name == ExternalVariablesModel.name, '\
                        'SlotsFiltersModel.external_variable_store_id == ExternalVariablesModel.store_id)')

    @property
    def type(self):
        return type(self).__factory__.get_filter_type(self.type_id)


class SlotsModelBase(AbstractConcreteBase):
    __tablename__ = 'slots'
    __swagger_json__ = get_swagger_json(__file__)

    id = sa.Column(sa.Integer, primary_key=True)
    max_items = sa.Column(sa.Integer, nullable=False)
    name = sa.Column(sa.String(255), nullable=False)

    @declared_attr
    def engine_id(cls):
        return sa.Column(sa.ForeignKey('engines.id'), nullable=False)

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def engine(cls):
        return sa.orm.relationship('EnginesModel')

    @declared_attr
    def slot_variables(cls):
        return sa.orm.relationship('SlotsVariablesModel', uselist=True, passive_deletes=True)

    @declared_attr
    def slot_filters(cls):
        return sa.orm.relationship('SlotsFiltersModel', uselist=True, passive_deletes=True)

    @declared_attr
    def fallbacks(cls):
        return sa.orm.relationship('SlotsModel',
                                   uselist=True, remote_side='SlotsModel.id',
                                   secondary='slots_fallbacks',
                                   primaryjoin='slots_fallbacks.c.slot_id == SlotsModel.id',
                                   secondaryjoin='slots_fallbacks.c.fallback_id == SlotsModel.id')

    async def init(self, session, input_=None, **kwargs):
        await super().init(session, input_=input_, **kwargs)
        self._validate_fallbacks(input_)
        self._validate_slot_variables(input_)
        self._validate_slot_filters(input_)

    def _validate_fallbacks(self, input_):
        for fallback in self.fallbacks:
            if fallback.id == self.id:
                raise SwaggerItModelError(
                    "a Engine Manager can't fallback itself", input_)

            if fallback.engine.item_type_id != self.engine.item_type_id:
                raise SwaggerItModelError(
                    "Cannot set a fallback with different items types", input_)

    def _validate_slot_variables(self, input_):
        if self.engine is not None:
            engine = self.engine.todict()
            engine_variables_set = set([var['name'] for var in engine['variables']])
            message = "Invalid slot variable with 'engine_variable_name' attribute value '{}'"
            schema = {'available_variables': engine['variables']}

            for slot_variable in self.slot_variables:
                var_name = slot_variable.engine_variable_name
                if var_name not in engine_variables_set:
                    raise ValidationError(
                        message.format(var_name),
                        instance=input_, schema=schema
                    )

    def _validate_slot_filters(self, input_):
        if self.engine is not None:
            engine = self.engine.todict()
            available_filters = engine['item_type']['available_filters']
            available_filters_set = set([filter_['name'] for filter_ in available_filters])
            schema = {'available_filters': available_filters}
            message = "Invalid slot filter with 'property_name' attribute value '{}'"

            for slot_filter in self.slot_filters:
                filter_name = slot_filter.property_name
                if filter_name not in available_filters_set:
                    raise ValidationError(
                        message.format(filter_name),
                        instance=input_, schema=schema
                    )

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'engine_id':
            value = {'id': value}
            attr_name = 'engine'

        if attr_name == 'slot_variables':
            for engine_var in value:
                if 'external_variable_id' in engine_var:
                    var = {'id': engine_var.pop('external_variable_id')}
                    engine_var['external_variable'] = var

        await super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('fallbacks') is not False:
            for fallback in dict_inst.get('fallbacks'):
                fallback.pop('fallbacks')


def build_slots_fallbacks_table(metadata, **kwargs):
    return sa.Table("slots_fallbacks", metadata,
        sa.Column("slot_id", sa.Integer, sa.ForeignKey(
            "slots.id", ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column("fallback_id", sa.Integer, sa.ForeignKey(
            "slots.id", ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import build_validator, get_module_path, set_logger
from myreco.engines.cores.utils import build_engine_data_path, build_engine_key_prefix, makedirs
from myreco.engines.cores.items_indices_map import ItemsIndicesMap
from jsonschema import Draft4Validator
from abc import ABCMeta


class EngineCoreMetaBase(ABCMeta):

    def __init__(cls, name, bases_classes, attributes):
        if hasattr(cls, '__configuration_schema__'):
            schema = cls.__configuration_schema__
            Draft4Validator.check_schema(schema)
            cls.__config_validator__ = build_validator(schema, get_module_path(cls))


class EngineCoreBase(metaclass=EngineCoreMetaBase):

    def __init__(self, engine, items_model):
        self.engine = engine
        self._items_model = items_model
        self._key = build_engine_key_prefix(self.engine)
        self._data_path = build_engine_data_path(self.engine)
        self._items_indices_map = ItemsIndicesMap(self._items_model)
        makedirs(self._data_path)
        set_logger(self, self._key)

    def validate_config(self):
        type(self).__config_validator__.validate(self.engine['configuration'])
        self._validate_config()

    def _validate_config(self):
        pass



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.items_types.model import ItemsTypesModelBase
from myreco.utils import extend_swagger_json
from swaggerit.exceptions import SwaggerItModelError
from jsonschema import Draft4Validator
from tempfile import NamedTemporaryFile
from aiofiles import zip_open, open as aiopen
from io import BytesIO
from collections import namedtuple
import ujson
import boto3
import os
import asyncio
import gc


class ItemsTypesDataFileImporterModelBase(ItemsTypesModelBase):
    __swagger_json__ = extend_swagger_json(
        ItemsTypesModelBase.__swagger_json__,
        __file__
    )

    @classmethod
    async def post_import_data_file_job(cls, req, session):
        content_type = req.headers.get('content-type')
        if content_type != 'application/zip':
            raise SwaggerItModelError("Invalid content type '{}'".format(content_type))

        stream = BytesIO()
        async for line in req.body:
            stream.write(line)
        stream.seek(0)

        return await cls._create_job(
            cls._run_import_data_file_job,
            req, session, '_importer', stream=stream)

    @classmethod
    async def _create_job(cls, func, req, session, jobs_id_prefix, stream=None):
        store_id = req.query['store_id']
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        session = cls._copy_session(session)
        jobs_id = store_items_model.__key__ + jobs_id_prefix
        return super()._create_job(
            func, jobs_id,
            req, session,
            store_items_model, store_id,
            stream=stream
        )

    @classmethod
    def _run_import_data_file_job(cls, req, session, store_items_model, store_id, stream):
        upload_file = req.query.get('upload_file', True)
        if upload_file:
            cls._put_file_on_s3(stream, store_items_model, session, store_id)
            stream.seek(0)

        result = asyncio.run_coroutine_threadsafe(
            cls._update_items_from_zipped_file(stream, store_items_model, session),
            session.loop
        ).result()

        gc.collect()
        return result

    @classmethod
    def _put_file_on_s3(cls, stream, store_items_model, session, store_id):
        cls._logger.info("Started put file on S3 for '{}'".format(store_items_model.__key__))

        store = asyncio.run_coroutine_threadsafe(
            cls.get_model('stores').get(session, [{'id': store_id}]),
            session.loop
        ).result()[0]

        s3_bucket = store['configuration']['aws']['s3']['bucket']
        access_key_id = store['configuration']['aws'].get('access_key_id')
        secret_access_key = store['configuration']['aws'].get('secret_access_key')
        s3_key = '{}.zip'.format(store_items_model.__key__)

        boto3.resource(
            's3',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key
        ).Bucket(s3_bucket).put_object(Body=stream, Key=s3_key)

        cls._logger.info("Finished put file on S3 for '{}'".format(store_items_model.__key__))

    @classmethod
    async def _update_items_from_zipped_file(cls, stream, store_items_model, session):
        feed = await cls._build_async_file(zip_open, stream)
        result = await cls._update_items_from_file(feed, store_items_model, session)

        await feed.close()
        os.remove(feed.name)
        return result

    @classmethod
    async def _build_async_file(cls, file_func, stream=None, mode=None):
        tempfile = NamedTemporaryFile(delete=False)
        filename = tempfile.name
        kwargs = {'mode': mode} if mode else {}
        if stream is not None:
            tempfile.write(stream.read())

        tempfile.close()
        af = await file_func(filename, **kwargs)
        af.name = filename
        return af

    @classmethod
    async def _update_items_from_file(cls, feed, store_items_model, session):
        cls._logger.info(
            "Started update items from file for '{}'".format(store_items_model.__key__)
        )
        warning_message = "Invalid line for model '{}': ".format(store_items_model.__key__) + '{}'
        validator = Draft4Validator(store_items_model.item_type['schema'])
        new_keys = set()
        success_lines = 0
        errors_lines = 0
        new_keys = set()
        success_lines = 0
        errors_lines = 0
        empty_lines = 0
        validator = Draft4Validator(store_items_model.item_type['schema'])
        new_keys = set()
        success_lines = 0
        errors_lines = 0
        empty_lines = 0
        old_keys = set(await session.redis_bind.hkeys(store_items_model.__key__))
        lines = []

        async for line in feed:
            try:
                line = cls._try_to_process_line(line, validator)
                if line is None:
                    empty_lines += 1
                    continue
                else:
                    success_lines += 1
                    new_keys.add(store_items_model.get_instance_key(line))
                    lines.append(line)

                    if len(lines) == 1000:
                        await store_items_model.insert(session, lines)
                        lines = []

            except Exception as error:
                errors_lines += 1
                cls._logger.warning(warning_message.format(line))
                cls._logger.warning(error)
                continue

        if lines:
            await store_items_model.insert(session, lines)

        del lines
        await feed.close()
        old_keys.difference_update(new_keys)

        if old_keys:
            await session.redis_bind.hdel(store_items_model.__key__, *old_keys)

        await cls._set_stock_filter(store_items_model, session)

        cls._logger.info(
            "Finished update items from file for '{}'".format(store_items_model.__key__)
        )

        return {
            'success_lines': success_lines,
            'errors_lines': errors_lines,
            'empty_lines': empty_lines
        }

    @classmethod
    def _try_to_process_line(cls, line, validator):
        line = line.strip().decode()
        if not line:
            return None

        line = ujson.loads(line)
        validator.validate(line)

        return line

    @classmethod
    async def get_import_data_file_job(cls, req, session):
        return await cls._get_job('_importer', req, session)

    @classmethod
    async def _get_job(cls, sufix, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        jobs_id = store_items_model.__key__ + sufix
        return await super()._get_job(jobs_id, req, session)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.factory import ModelsFactory
from myreco.api import MyrecoAPI
from swaggerit.models.orm.session import Session
from swaggerit.models.orm.binds import ElSearchBind
from unittest import mock
from sqlalchemy import create_engine
from aioredis import create_redis
from base64 import b64encode
from sqlalchemy.schema import DDL
import sqlalchemy as sa
import pytest
import pymysql
import asyncio
import uvloop


@pytest.fixture(scope='session')
def base_model(api):
    return api.models_factory.base_model


@pytest.fixture(scope='session')
def models(api):
    return api.all_models


@pytest.fixture(scope='session')
def loop():
    loop = uvloop.new_event_loop()
    # loop = asyncio.new_event_loop() # just for debugging
    asyncio.set_event_loop(loop)
    return loop


@pytest.fixture(scope='session')
def redis(variables, loop):
    coro = create_redis(
        (variables['redis']['host'], variables['redis']['port']),
        db=variables['redis']['db'],
        loop=loop
    )
    return loop.run_until_complete(coro)


@pytest.fixture(scope='session')
def pymysql_conn(variables):
    database = variables['database'].pop('database')
    conn = pymysql.connect(**variables['database'])

    with conn.cursor() as cursor:
        try:
            cursor.execute('drop database {};'.format(database))
        except:
            pass
        cursor.execute('create database {};'.format(database))
        cursor.execute('use {};'.format(database))
    conn.commit()
    variables['database']['database'] = database

    return conn


@pytest.fixture(scope='session')
def engine(variables, pymysql_conn):
    if variables['database']['password']:
        url = 'mysql+pymysql://{user}:{password}'\
            '@{host}:{port}/{database}'.format(**variables['database'])
    else:
        variables['database'].pop('password')
        url = 'mysql+pymysql://{user}'\
            '@{host}:{port}/{database}'.format(**variables['database'])
        variables['database']['password'] = None

    return create_engine(url)


@pytest.fixture(scope='session')
def elsearch(variables, loop):
    es = ElSearchBind(**variables['elsearch'])
    loop.run_until_complete(es.create_index())
    return es


@pytest.fixture(scope='session')
def api(engine, redis, elsearch, loop):
    api = MyrecoAPI(sqlalchemy_bind=engine, redis_bind=redis,
                    elsearch_bind=elsearch, title='Myreco API',
                    loop=loop, debug=True, type_='objects_exporter')

    class ModelTest(api.models_factory.base_model):
        __tablename__ = 'test'
        id = sa.Column(sa.Integer, primary_key=True)

        @classmethod
        async def do_nothing(cls, req, resp, **kwargs):
            return cls._build_response(200)

        __swagger_json__ = {
            'paths': {
                '/testing': {
                    'parameters': [{
                        'name': 'Authorization',
                        'in': 'header',
                        'required': True,
                        'type': 'string'
                    }],
                    'post': {
                        'operationId': 'do_nothing',
                        'responses': {'201': {'description': 'Created'}}
                    }
                },
                '/testing/{id}': {
                    'parameters': [{
                        'name': 'id',
                        'in': 'path',
                        'required': True,
                        'type': 'integer'
                    },{
                        'name': 'Authorization',
                        'in': 'header',
                        'required': True,
                        'type': 'string'
                    }],
                    'post': {
                        'operationId': 'do_nothing',
                        'responses': {'201': {'description': 'Created'}}
                    }
                }
            }
        }

    api.add_model(ModelTest)
    api.router._resources = api.router._resources[-8:] + api.router._resources[:-8]
    return api


@pytest.fixture
def client(api, test_client):
    return test_client(api)


@pytest.fixture
def session(variables, redis, engine, pymysql_conn, base_model, loop, elsearch):
    base_model.metadata.bind = engine
    base_model.metadata.create_all()

    with pymysql_conn.cursor() as cursor:
        cursor.execute('SET FOREIGN_KEY_CHECKS = 0;')
        for table in base_model.metadata.tables.values():
            cursor.execute('delete from {};'.format(table))

            try:
                cursor.execute('alter table {} auto_increment=1;'.format(table))
            except:
                pass
        cursor.execute('SET FOREIGN_KEY_CHECKS = 1;')
        cursor.execute('ALTER TABLE slots_filters CHANGE id id INT AUTO_INCREMENT;')
        cursor.execute('ALTER TABLE slots_variables CHANGE id id INT AUTO_INCREMENT;')

    pymysql_conn.commit()
    loop.run_until_complete(redis.flushdb())
    loop.run_until_complete(elsearch.flush_index())
    session = Session(bind=engine, redis_bind=redis, elsearch_bind=elsearch, loop=loop)
    yield session
    session.close()


@pytest.fixture
def headers():
    return {
        'Authorization': 'Basic ' + b64encode('test:test'.encode()).decode(),
        'Content-Type': 'application/json'
    }


@pytest.fixture
def headers_without_content_type():
    return {
        'Authorization': 'Basic ' + b64encode('test:test'.encode()).decode()
    }



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.filters.filters import (BooleanFilterBy, SimpleFilterBy, ObjectFilterBy,
    ArrayFilterBy, SimpleFilterOf, ObjectFilterOf,
    ArrayFilterOf, IndexFilterOf, IndexFilterByPropertyOf)


class FiltersFactory(object):
    _filters_types_map = {
        'property_value': {
            'name': 'Filter by property value',
            'types': {
                'integer': SimpleFilterBy,
                'string': SimpleFilterBy,
                'object': ObjectFilterBy,
                'array': ArrayFilterBy,
                'boolean': BooleanFilterBy
            }
        },
        'item_property_value': {
            'name': 'Filter by item property value',
            'types': {
                'integer': SimpleFilterOf,
                'string': SimpleFilterOf,
                'object': ObjectFilterOf,
                'array': ArrayFilterOf,
                'boolean': BooleanFilterBy
            }
        },
        'property_value_index': {
            'name': 'Filter by property value index',
            'types': {
                'integer': IndexFilterOf,
                'string': IndexFilterOf,
                'object': IndexFilterOf,
                'array': IndexFilterOf,
                'boolean': IndexFilterOf
            }
        },
        'item_property_value_index': {
            'name': 'Filter by item property value index',
            'types': {
                'integer': IndexFilterByPropertyOf,
                'string': IndexFilterByPropertyOf,
                'object': IndexFilterByPropertyOf,
                'array': IndexFilterByPropertyOf,
                'boolean': IndexFilterByPropertyOf
            }
        }
    }

    @classmethod
    def get_filter_types(cls):
        return [{'name': filter_type['name'], 'id': filter_type_id}
            for filter_type_id, filter_type in cls._filters_types_map.items()]

    @classmethod
    def get_filter_type(cls, filter_type_id):
        filter_type = cls._filters_types_map.get(filter_type_id)
        return {
            'name': filter_type['name'],
            'id': filter_type_id
        } if filter_type else None

    @classmethod
    def make(cls, items_model, slot_filter, schema, skip_values=None):
        value_type = schema['type']
        filter_name = slot_filter['property_name']
        type_id = slot_filter['type_id']
        is_inclusive = slot_filter['is_inclusive']
        id_names = schema.get('id_names')
        filter_class = cls._filters_types_map.get(type_id, {'types': {}})['types'].get(value_type)

        if filter_class:
            return filter_class(items_model, filter_name, is_inclusive, id_names, skip_values)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.base import EngineCoreBase
from myreco.engines.exceptions import EngineError
from abc import abstractmethod
from swaggerit.json_builder import JsonBuilder
from numpy import argpartition


class EngineCoreRecommender(EngineCoreBase):

    def get_variables(self):
        return []

    async def get_items(self, session, filters, max_items, show_details, **external_variables):
        items_vector = await self._build_items_vector(session, **external_variables)

        if items_vector is not None:
            for filter_, ids in filters.items():
                await filter_.filter(session, items_vector, ids)

            return await self._build_rec_list(session, items_vector, max_items, show_details)

        return []

    @abstractmethod
    async def _build_items_vector(self, session, **external_variables):
        pass

    async def _build_rec_list(self, session, items_vector, max_items, show_details):
        best_indices = self._get_best_indices(items_vector, max_items)
        best_items_keys = await self._items_indices_map.get_items(best_indices, session)

        if show_details and best_items_keys:
            return await self._items_model.get(session, best_items_keys)

        else:
            items_ids = []
            for key in best_items_keys:
                item = {}
                self._items_model.set_instance_ids(item, key)
                self._set_item_values(item)
                items_ids.append(item)

            return items_ids

    def _get_best_indices(self, items_vector, max_items):
        if max_items > items_vector.size:
            max_items = items_vector.size

        best_indices = argpartition(-items_vector, max_items-1)[:max_items]
        best_values = items_vector[best_indices]
        return [int(i) for i, v in
            sorted(zip(best_indices, best_values), key=lambda x: x[1], reverse=True) if v > 0.0]

    def _set_item_values(self, item):
        for k in item:
            schema = self.engine['item_type']['schema']['properties'].get(k)
            if schema is None:
                raise EngineError('Invalid Item {}'.format(item))

            item[k] = JsonBuilder.build(item[k], schema)
